package Pinwheel;

our $VERSION = '0.2.7';

1;

__DATA__

=head1 NAME

Pinwheel - A Rails-like framework for building dynamic web applications in Perl.

=head1 DESCRIPTION

TODO.

=head1 TESTS

The tests require access to a mysql database, the details of which are
hard-wired: host=localhost, database=framework_test, user=root, password="".

It suffices to create "framework_test" as an empty database prior to running
the tests.

=head1 AUTHOR

A&M Network Publishing <DLAMNetPub@bbc.co.uk>

=cut
